@@ -5,7 +5,7 @@ import logging
 import sys
 import typing
 from types import TracebackType
-from urllib.parse import SplitResult, parse_qsl, urlsplit, unquote
+from urllib.parse import SplitResult, parse_qsl, unquote, urlsplit
 
 from sqlalchemy import text
 from sqlalchemy.sql import ClauseElement
@@ -14,9 +14,9 @@ from databases.importer import import_from_string
 from databases.interfaces import ConnectionBackend, DatabaseBackend, TransactionBackend
 
 if sys.version_info >= (3, 7):  # pragma: no cover
-    from contextvars import ContextVar
+    import contextvars as contextvars
 else:  # pragma: no cover
-    from aiocontextvars import ContextVar
+    import aiocontextvars as contextvars
 
 try:  # pragma: no cover
     import click
@@ -46,6 +46,7 @@ class Database:
         "postgresql+aiopg": "databases.backends.aiopg:AiopgBackend",
         "postgres": "databases.backends.postgres:PostgresBackend",
         "mysql": "databases.backends.mysql:MySQLBackend",
+        "mysql+asyncmy": "databases.backends.asyncmy:AsyncMyBackend",
         "sqlite": "databases.backends.sqlite:SQLiteBackend",
     }
 
@@ -62,13 +63,15 @@ class Database:
 
         self._force_rollback = force_rollback
 
-        backend_str = self.SUPPORTED_BACKENDS[self.url.scheme]
+        backend_str = self._get_backend()
         backend_cls = import_from_string(backend_str)
         assert issubclass(backend_cls, DatabaseBackend)
         self._backend = backend_cls(self.url, **self.options)
 
         # Connections are stored as task-local state.
-        self._connection_context = ContextVar("connection_context")  # type: ContextVar
+        self._connection_context = contextvars.ContextVar(
+            "connection_context"
+        )  # type: contextvars.ContextVar
 
         # When `force_rollback=True` is used, we use a single global
         # connection, within a transaction that always rolls back.
@@ -79,7 +82,9 @@ class Database:
         """
         Establish the connection pool.
         """
-        assert not self.is_connected, "Already connected."
+        if self.is_connected:
+            logger.debug("Already connected, skipping connection")
+            return None
 
         await self._backend.connect()
         logger.info(
@@ -102,7 +107,9 @@ class Database:
         """
         Close all connections in the connection pool.
         """
-        assert self.is_connected, "Already disconnected."
+        if not self.is_connected:
+            logger.debug("Already disconnected, skipping disconnection")
+            return None
 
         if self._force_rollback:
             assert self._global_connection is not None
@@ -112,6 +119,8 @@ class Database:
 
             self._global_transaction = None
             self._global_connection = None
+        else:
+            self._connection_context = contextvars.ContextVar("connection_context")
 
         await self._backend.disconnect()
         logger.info(
@@ -173,6 +182,11 @@ class Database:
             async for record in connection.iterate(query, values):
                 yield record
 
+    def _new_connection(self) -> "Connection":
+        connection = Connection(self._backend)
+        self._connection_context.set(connection)
+        return connection
+
     def connection(self) -> "Connection":
         if self._global_connection is not None:
             return self._global_connection
@@ -180,14 +194,23 @@ class Database:
         try:
             return self._connection_context.get()
         except LookupError:
-            connection = Connection(self._backend)
-            self._connection_context.set(connection)
-            return connection
+            return self._new_connection()
 
     def transaction(
         self, *, force_rollback: bool = False, **kwargs: typing.Any
     ) -> "Transaction":
-        return Transaction(self.connection, force_rollback=force_rollback, **kwargs)
+        try:
+            connection = self._connection_context.get()
+            is_root = not connection._transaction_stack
+            if is_root:
+                newcontext = contextvars.copy_context()
+                get_conn = lambda: newcontext.run(self._new_connection)
+            else:
+                get_conn = self.connection
+        except LookupError:
+            get_conn = self.connection
+
+        return Transaction(get_conn, force_rollback=force_rollback, **kwargs)
 
     @contextlib.contextmanager
     def force_rollback(self) -> typing.Iterator[None]:
@@ -198,6 +221,11 @@ class Database:
         finally:
             self._force_rollback = initial
 
+    def _get_backend(self) -> str:
+        return self.SUPPORTED_BACKENDS.get(
+            self.url.scheme, self.SUPPORTED_BACKENDS[self.url.dialect]
+        )
+
 
 class Connection:
     def __init__(self, backend: DatabaseBackend) -> None:
@@ -215,8 +243,12 @@ class Connection:
     async def __aenter__(self) -> "Connection":
         async with self._connection_lock:
             self._connection_counter += 1
-            if self._connection_counter == 1:
-                await self._connection.acquire()
+            try:
+                if self._connection_counter == 1:
+                    await self._connection.acquire()
+            except Exception as e:
+                self._connection_counter -= 1
+                raise e
         return self
 
     async def __aexit__(
@@ -441,7 +473,11 @@ class DatabaseURL:
 
     @property
     def hostname(self) -> typing.Optional[str]:
-        return self.components.hostname
+        return (
+            self.components.hostname
+            or self.options.get("host")
+            or self.options.get("unix_sock")
+        )
 
     @property
     def port(self) -> typing.Optional[int]:
